(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                 Boite  outils pour les manipulations d'arbres         |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)
(* Michel Quercia, le 10/01/98 *)


                (* +-----------------------------------------+
                   |  Types des arbres gnraux et binaires  |
                   +-----------------------------------------+ *)

type 'a g_arbre = G_noeud of 'a * ('a g_arbre list);;
type 'a b_arbre = B_vide | B_noeud of 'a * ('a b_arbre) * ('a b_arbre);;


                     (* +-------------------------------+
                        |  Fonctions de transformation  |
                        +-------------------------------+ *)

value g_map : ('a -> 'b) -> ('a g_arbre -> 'b g_arbre);;
value b_map : ('a -> 'b) -> ('a b_arbre -> 'b b_arbre);;
(* [x_map f a] applique [f]  chaque noeud de [a] et renvoie *)
(* l'arbre des rsultats.                                    *)

                     (* +------------------------------+
                        |  Fonctions de mise en arbre  |
                        +------------------------------+ *)

value string_g_arbre : string -> string g_arbre;;
value string_b_arbre : string -> string b_arbre;;
value int_g_arbre    : string -> int    g_arbre;;
value int_b_arbre    : string -> int    b_arbre;;
(* [x_arbre chaine] interprte [chaine] comme la description parenthse *)
(* d'un arbre et renvoie cet arbre.                                      *)


                       (* +--------------------------+
                          |  Fonctions d'impression  |
                          +--------------------------+ *)

value string_of_string_g_arbre : string g_arbre -> string;;
value string_of_string_b_arbre : string b_arbre -> string;;
value string_of_int_g_arbre    : int    g_arbre -> string;;
value string_of_int_b_arbre    : int    b_arbre -> string;;
(* conversion arbre -> expression parenthse. *)

value print_string_g_arbre : string g_arbre -> unit;;
value print_string_b_arbre : string b_arbre -> unit;;
value print_int_g_arbre    : int    g_arbre -> unit;;
value print_int_b_arbre    : int    b_arbre -> unit;;
(* impression semi-graphique avec la bibliothque [format]. *)


(* +------------------------------------------------------------------------+
   |                                                                        |
   |                       Evaluation de formules infixes                   |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

type 'a token =
| Valeur   of 'a                      (* valeur reconnues                *)
| Prfixe  of 'a -> 'a                (* oprateur prfixe               *)
| Postfixe of 'a -> 'a                (* oprateur postfixe              *)
| Infixe   of int * ('a -> 'a -> 'a)  (* priorit, oprateur infixe      *)
| InfPref  of (int*('a->'a->'a)) * ('a->'a) (* infixe ou prfixe         *)
| Liste    of 'a list -> 'a           (* oprateur prfixe sur une liste *)
| Lpar     			      (* parenthse gauche		 *)
| Rpar     			      (* parenthse droite		 *)
| Lsep                                (* sparateur de liste             *)
;;

exception Erreur of exn * int;;
value print_erreur : string -> int -> unit;;

(* [print_erreur msg s i] affiche sur la sortie standard la chaine [s]
   et place un accent circonflexe sous le caractre numro [i] de [s].
*)

value analyseur : (unit -> (int * ('b -> 'b -> 'b))) ->
                  ('a -> 'b token) ->
                  (char stream -> 'a stream) ->
                  string -> 'b;;

(* [analyseur impl f lexer s] analyse la chaine [s] et renvoie la valeur
   qu'elle reprsente. [s] est d'abord convertie en flux de caractres, ce
   flux est soumis  l'analyseur lexical [lexer], puis les lexmes obtenus
   sont convertis en lments syntaxiques par [f].
   [impl] retourne un couple (priorit, opration infixe implicite) ou
   dclenche [Parse_failure] s'il n'y a pas d'opration implicite.

   L'analyseur proprement dit value le flux d'lments syntaxiques dans
   l'ordre suivant :

     - valuation des expressions entre parenthses ;
     - valuation des oprateurs prfixes ;
     - valuation des oprateurs postfixes ;
     - valuation des oprateurs infixes par priorit dcroissante,
       a op b op' c est valu comme (a op b) op' c lorsque op et op'
       ont mme priorit.

   Le code de l'analyseur suppose que tous les oprateurs infixe ont une
   priorit suprieure ou gale  1, la priorit 0 est rserve pour usage
   interne.

   En cas d'erreur de syntaxe ou de calcul, il est gnr une exception
   [Erreur(ex,i)] o [ex] est l'exception dclenche, et [i] est la position
   du caractre ayant dclench cette exception.

   Par exemple, le code suivant fournit un analyseur des expressions
   arithmtiques :

   -------------------------------------------------------------------------
   #open "genlex";;
   #open "arbres";;
   load_object "arbres";;

   let value = 
     let lexeur = make_lexer ["+"; "-"; "*"; "/"; ""; "plus"; "mult"; "("; ")"; ";"]
     and syntax_of_lexme = function
         | Kwd "+" -> Infixe(1,fun x y -> x+y)
         | Kwd "-" -> InfPref((1,fun x y -> x-y),(fun x -> -x))
         | Kwd "*" -> Infixe(2,fun x y -> x*y)
         | Kwd "/" -> Infixe(2,fun x y -> x/y)
         | Kwd "" -> Postfixe(fun x -> x*x)
         | Kwd "plus" -> Liste(fun x -> it_list (prefix +) 0 x)
         | Kwd "mult" -> Liste(fun x -> it_list (prefix * ) 1 x)
         | Int n -> Valeur n
         | Kwd "(" -> Lpar
         | Kwd ")" -> Rpar
         | Kwd ";" -> Lsep
         | _       -> raise Parse_error
     in let eval = analyseur (fun () -> raise Parse_failure) syntax_of_lexme lexeur
     in function s -> try eval s
                      with Erreur(ex,i) -> print_erreur s i; raise ex
   ;;

   value "2 + 4*plus(1; 2; 3)/(-(8-2))";;  (-> -22)
   -------------------------------------------------------------------------

   le code suivant fournit un analyseur d'expressions algbriques :

   -------------------------------------------------------------------------

   #open "genlex";;
   #open "arbres";;
   load_object "arbres";;

   type 'a expression =
   | Const of 'a
   | Var   of string
   | Fct   of string * 'a expression
   | Bin   of binop * 'a expression * 'a expression
   and binop = Plus | Moins | Mult | Div;;

   let value fcts = 
     let lexeur = make_lexer ["+"; "-"; "*"; "/"; "("; ")" ]
     and syntax_of_lexme = function
         | Kwd "+"  -> Infixe(1,fun x y -> Bin(Plus, x,y))
         | Kwd "-"  -> InfPref((1,fun x y -> Bin(Moins,x,y)),(fun x -> Fct("opp",x)))
         | Kwd "*"  -> Infixe(2,fun x y -> Bin(Mult, x,y))
         | Kwd "/"  -> Infixe(2,fun x y -> Bin(Div,  x,y))
         | Int n    -> Valeur(Const n)
         | Ident i  -> if mem i fcts then Prfixe(fun x -> Fct(i,x))
                                     else Valeur(Var i)
         | Kwd "("  -> Lpar
         | Kwd ")"  -> Rpar
         | _        -> raise Parse_error
     in let eval = analyseur (fun () -> (2,fun x y -> Bin(Mult,x,y)))
                             syntax_of_lexme lexeur
     in function s -> try eval s
                      with Erreur(ex,i) -> print_erreur s i; raise ex
   ;;

   value ["sin"; "cos"; "exp"] "-x + 2y/sin(x+2)";; ->
   Bin(Plus, Fct("opp",Var "x"),
             Bin(Div, Bin(Mult, Const 2,Var "y"),
                      Fct("sin", Bin(Plus, Var "x" Const 2))))

   -------------------------------------------------------------------------

   Les expressions peuvent tre affiches sous forme arborescente :

   -------------------------------------------------------------------------

   let rec string_g_arbre_of_expression = function
   | Const(n) -> G_noeud(string_of_int n,[])
   | Var(v)   -> G_noeud(v,[])
   | Fct(f,e) -> G_noeud(f,[string_g_arbre_of_expression e])
   | Bin(op,e,e') -> G_noeud(
      (match op with Plus -> "+" | Moins -> "-" | Mult -> "*" | Div -> "/"),
      [string_g_arbre_of_expression e; string_g_arbre_of_expression e'])
   ;;

   let print_expression(e) =
     print_string_g_arbre(string_g_arbre_of_expression e)
   ;;

   install_printer "print_expression";;

   -------------------------------------------------------------------------

*)
