(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                               Manipulation d'automates                 |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

(* Michel Quercia, le 10/01/98 *)

#open "graphes";;

(* automate gnrique  tats de type 'a, tiquettes de type 'b  et lettres de type 'c *)
type ('a,'b,'c) automate = {
  g  : ('a,'b,'c lettre list) graphe;   (* graphe de transition *)
  initial : 'a list;                    (* tats de dpart      *)
  final   : 'a list                     (* tats d'acceptation  *)
}
and 'a lettre = Epsilon | Lettre of 'a
;;

                 (* +----------------------------+
                    |  Construction d'automates  |
                    +----------------------------+ *)

(* l'automate fourni en argument n'est pas modifi, un nouvel automate est retourn. *)

value new_automate : unit -> ('a,'b,'c) automate
  and add_tat     : ('a,'b,'c) automate -> 'a -> 'b -> ('a,'b,'c) automate
  and del_tat     : ('a,'b,'c) automate -> 'a       -> ('a,'b,'c) automate
  and add_trans    : ('a,'b,'c) automate -> 'a*'a -> 'c lettre -> ('a,'b,'c) automate
  and del_trans    : ('a,'b,'c) automate -> 'a*'a -> 'c lettre -> ('a,'b,'c) automate
;;
(*
 [new_automate ()] retourne un automate sans tats ni transitions.
 [add_tat a e x]  ajoute l'tat [e] d'tiquette [x].
 [del_tat a e]    supprime l'tat [e] et les transitions associes.
 [add_trans a (x,y) c] ajoute une transition de [x] vers [y] sur [c], [c] est
                       une lettre ou [Epsilon].
 [del_trans a (x,y) c] supprime la transition de [x] vers [y] sur [c]. les
                       autres transitions de [x] vers [y] sont conserves.
*)

value mirroir  : ('a,'b,'c) automate -> ('a,'b,'c) automate;;
(* Automate mirroir *)


value map_auto : (('a*'b) -> ('aa*'bb)) -> ('a,'b,'c) automate -> ('aa,'bb,'c) automate
  and map_int  : ('a,'b,'c) automate -> (int,string,'c) automate;;
(*
 [map_auto f a] renomme les tats de l'automate [a]. La fonction [f] est
 appele une unique fois pour chaque tat de [a] et reoit en argument
 un couple (tat,tiquette), elle doit retourner un couple (tat',tiquette').
 Le comportement de [map_auto] n'est pas spcifi si [f] n'est pas injective.

 [map_int a] retourne un automate quivalent  [a]  tats entiers.
 L'tiquette d'un entier est sa reprsentation dcimale.
*)

                        (* +--------------+
                           |  Simulation  |
                           +--------------+ *)

value cltures : ('a,'b,'c) automate -> ('a,'a list) map__t;;
(* Calcule les cltures par transitions spontanes des tats de [a] *)

value avance : ('a,'b,'c) automate ->
               ('a,'a list) map__t ->
               'a list ->
               'c lettre -> 'a list;;
(*
 [avance a cl tats c] calcule la liste des tats dduits de [tats] par
 la transition sur [c]. [a] est l'automate et [cl] la table des cltures
 retourne par [cltures]. La liste retourne est trie sans rptition et est
 ferme par epsilon-transitions. On obtient la clture par epsilon-transitions
 d'une liste d'tats par : [avance a cl tats Epsilon].
*)


      (* +---------------------------------------------------+
         |  expression rgulire sur les lettres de type 'a  |
         +---------------------------------------------------+ *)

type 'a expression =
| Vide
| Const  of 'a lettre
| Union  of 'a expression list
| Concat of 'a expression list
| Etoile of 'a expression
| Plus   of 'a expression
;;

value expression : string -> char expression;;
(*
 Analyse syntaxique et mise en arbre d'une expression rgulire.
 Caractres spciaux :

    @      -> Epsilon
    <>     -> langage vide
    .      -> concatnation
    |      -> somme
    *      -> toile
    +      -> x+ <=> x.x*
    ()     -> regroupement
    espace -> ignor
    \...   -> squence d'chappement dsignant un caractre ne devant pas tre
              interprt. La squence \nnn dsigne le caractre de code dcimal
              nnn, les autres squences traditionnelles sont reconnues.

 La concatnation est implicite, elle est prioritaire sur la somme.
 Les oprations postfixes ( + et * ) sont prioritaires sur la concatnation.
*)

value print_expression : char expression -> unit;;
(* Impression semi-graphique  l'aide de la bibliothque [format]. *)

value automate_of_expression : 'a expression -> (int,string,'a) automate;;
(* Automate de Thompson associ  une expression rgulire. *)


                (* +------------------------------+
                   |  Transformation d'automates  |
                   +------------------------------+ *)

value simplifie : ('a,'b,'c) automate -> ('a,'b,'c) automate;;
(*
 Retourne un automate quivalent en court-circuitant les tats d'o ne
 partent que des transitions spontanes ou n'arrivent que des transitions
 spontanes. Supprime aussi les transitions spontanes d'un tat vers
 lui-mme. La simplification n'est garantie complte que si tous les tats
 sont accessibles et co-accessibles, sinon il peut tre ncessaire
 de procder  plusieurs simplifications d'affile.
*)

value dterminise : ('a,string,'c) automate -> ('a list,string,'c) automate
  and minimise    : ('a,'b,'c) automate -> (int,string,'c) automate;;
(*
 [dterminise a] retourne un automate dterministe quivalent. L'tiquette
                 d'une liste d'tats est obtenue par concatnation des
                 tiquettes des tats de la liste.
 [minimise a]    retourne un automate dterministe minimal quivalent.
*)



                          (* +----------+
                             |  Divers  |
                             +----------+ *)


value output_automate : out_channel -> ('a,'b,'c) automate -> unit
  and input_automate  : in_channel  -> ('a,'b,'c) automate
;;
(*
 criture sur ou relecture depuis un canal d'entre/sortie d'un automate.
 Le type de l'automate relu depuis un canal d'entre doit correspondre 
 son utilisation, aucun contrle de type n'est effectu.
*)

                  (* +---------------------------+
                     |  Affichage et annimation  |
                     +---------------------------+ *)

#open "graphx";;

value xgraphe_of_automate : ('a,string,char) automate -> 'a xgraphe;;
(* calcule les positions des tats, des arcs et des lgendes. *)

value show  : ('a,string,char) automate -> 'a xgraphe;;
(*
 Affiche un automate dans la fentre graphique et permet le dplacement des
 tats, des arcs et des lgendes, la sortie dans un fichier postscript et
 la simulation. Appuyer sur <Esc> pour arrter la simulation, sur <BS> pour
 revenir en arrire.
*)

value xshow : ('a,string,char) automate -> unit;;
(*
 Identique  [show], mais lance un processus autonome pour afficher
 l'automate dans une fentre indpendante. Le programme [showauto]
 doit tre accessible par la variable $PATH.
*)
